@@ -1,43 +1,51 @@
-#include	"ruby/config.h"
+#include "ruby/config.h"
+
 #ifdef RUBY_EXTCONF_H
-#include RUBY_EXTCONF_H
+# include RUBY_EXTCONF_H
 #endif
-#include	<stdlib.h>
-#include	<stdio.h>
-#include	<sys/types.h>
-#include	<sys/stat.h>
-#include	<sys/file.h>
-#include	<fcntl.h>
-#include	<errno.h>
-#include	<pwd.h>
+
+#include <ctype.h>
+#include <errno.h>
+#include <signal.h>
+#include <stdio.h>
+#include <stdlib.h>
+
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <sys/file.h>
+#include <fcntl.h>
+
+#ifdef HAVE_PWD_H
+# include <pwd.h>
+#endif
+
 #ifdef HAVE_SYS_IOCTL_H
-#include	<sys/ioctl.h>
+# include <sys/ioctl.h>
 #endif
+
 #ifdef HAVE_LIBUTIL_H
-#include	<libutil.h>
+# include <libutil.h>
 #endif
+
 #ifdef HAVE_UTIL_H
-#include	<util.h>
+# include <util.h>
 #endif
+
 #ifdef HAVE_PTY_H
-#include	<pty.h>
+# include <pty.h>
 #endif
+
 #if defined(HAVE_SYS_PARAM_H)
-  /* for __FreeBSD_version */
+ /* for __FreeBSD_version */
 # include <sys/param.h>
 #endif
+
 #ifdef HAVE_SYS_WAIT_H
-#include <sys/wait.h>
+# include <sys/wait.h>
 #else
-#define WIFSTOPPED(status)    (((status) & 0xff) == 0x7f)
+# define WIFSTOPPED(status) (((status) & 0xff) == 0x7f)
 #endif
-#include <ctype.h>
 
-#include "internal.h"
-#include "ruby/io.h"
-#include "ruby/util.h"
-
-#include <signal.h>
 #ifdef HAVE_SYS_STROPTS_H
 #include <sys/stropts.h>
 #endif
@@ -46,6 +54,12 @@
 #include <unistd.h>
 #endif
 
+#include "internal.h"
+#include "internal/process.h"
+#include "internal/signal.h"
+#include "ruby/io.h"
+#include "ruby/util.h"
+
 #define	DEVICELEN	16
 
 #ifndef HAVE_SETEUID
@@ -78,9 +92,13 @@ struct pty_info {
 
 static void getDevice(int*, int*, char [DEVICELEN], int);
 
+static int start_new_session(char *errbuf, size_t errbuf_len);
+static int obtain_ctty(int master, int slave, const char *slavename, char *errbuf, size_t errbuf_len);
+static int drop_privilige(char *errbuf, size_t errbuf_len);
+
 struct child_info {
     int master, slave;
-    char *slavename;
+    const char *slavename;
     VALUE execarg_obj;
     struct rb_execarg *eargp;
 };
@@ -88,26 +106,42 @@ struct child_info {
 static int
 chfunc(void *data, char *errbuf, size_t errbuf_len)
 {
-    struct child_info *carg = data;
+    const struct child_info *carg = data;
     int master = carg->master;
     int slave = carg->slave;
+    const char *slavename = carg->slavename;
+
+    if (start_new_session(errbuf, errbuf_len))
+        return -1;
+
+    if (obtain_ctty(master, slave, slavename, errbuf, errbuf_len))
+        return -1;
+
+    if (drop_privilige(errbuf, errbuf_len))
+        return -1;
+
+    return rb_exec_async_signal_safe(carg->eargp, errbuf, errbuf_len);
+}
 
 #define ERROR_EXIT(str) do { \
-	strlcpy(errbuf, (str), errbuf_len); \
-	return -1; \
+        strlcpy(errbuf, (str), errbuf_len); \
+        return -1; \
     } while (0)
 
-    /*
-     * Set free from process group and controlling terminal
-     */
+/*
+ * Set free from process group and controlling terminal
+ */
+static int
+start_new_session(char *errbuf, size_t errbuf_len)
+{
 #ifdef HAVE_SETSID
     (void) setsid();
 #else /* HAS_SETSID */
 # ifdef HAVE_SETPGRP
-#  ifdef SETGRP_VOID
+#  ifdef SETPGRP_VOID
     if (setpgrp() == -1)
         ERROR_EXIT("setpgrp()");
-#  else /* SETGRP_VOID */
+#  else /* SETPGRP_VOID */
     if (setpgrp(0, getpid()) == -1)
         ERROR_EXIT("setpgrp()");
     {
@@ -118,20 +152,25 @@ chfunc(void *data, char *errbuf, size_t errbuf_len)
             ERROR_EXIT("ioctl(TIOCNOTTY)");
         close(i);
     }
-#  endif /* SETGRP_VOID */
+#  endif /* SETPGRP_VOID */
 # endif /* HAVE_SETPGRP */
 #endif /* HAS_SETSID */
+    return 0;
+}
 
-    /*
-     * obtain new controlling terminal
-     */
+/*
+ * obtain new controlling terminal
+ */
+static int
+obtain_ctty(int master, int slave, const char *slavename, char *errbuf, size_t errbuf_len)
+{
 #if defined(TIOCSCTTY)
     close(master);
     (void) ioctl(slave, TIOCSCTTY, (char *)0);
     /* errors ignored for sun */
 #else
     close(slave);
-    slave = rb_cloexec_open(carg->slavename, O_RDWR, 0);
+    slave = rb_cloexec_open(slavename, O_RDWR, 0);
     if (slave < 0) {
         ERROR_EXIT("open: pty slave");
     }
@@ -141,46 +180,56 @@ chfunc(void *data, char *errbuf, size_t errbuf_len)
     dup2(slave,0);
     dup2(slave,1);
     dup2(slave,2);
-    close(slave);
+    if (slave < 0 || slave > 2) (void)!close(slave);
+    return 0;
+}
+
+static int
+drop_privilige(char *errbuf, size_t errbuf_len)
+{
 #if defined(HAVE_SETEUID) || defined(HAVE_SETREUID) || defined(HAVE_SETRESUID)
     if (seteuid(getuid())) ERROR_EXIT("seteuid()");
 #endif
-
-    return rb_exec_async_signal_safe(carg->eargp, errbuf, sizeof(errbuf_len));
-#undef ERROR_EXIT
+    return 0;
 }
 
+#undef ERROR_EXIT
+
 static void
 establishShell(int argc, VALUE *argv, struct pty_info *info,
-	       char SlaveName[DEVICELEN])
+               char SlaveName[DEVICELEN])
 {
     int 		master, slave, status = 0;
     rb_pid_t		pid;
-    char		*p, *getenv();
-    struct passwd	*pwent;
+    char		*p;
     VALUE		v;
     struct child_info   carg;
     char		errbuf[32];
 
     if (argc == 0) {
-	const char *shellname;
+        const char *shellname = "/bin/sh";
 
-	if ((p = getenv("SHELL")) != NULL) {
-	    shellname = p;
-	}
-	else {
-	    pwent = getpwuid(getuid());
-	    if (pwent && pwent->pw_shell)
-		shellname = pwent->pw_shell;
-	    else
-		shellname = "/bin/sh";
-	}
-	v = rb_str_new2(shellname);
-	argc = 1;
-	argv = &v;
+        if ((p = getenv("SHELL")) != NULL) {
+            shellname = p;
+        }
+        else {
+#if defined HAVE_PWD_H
+            const char *username = getenv("USER");
+            if (username == NULL)
+                username = getlogin();
+            if (username != NULL) {
+                struct passwd *pwent = getpwnam(username);
+                if (pwent && pwent->pw_shell)
+                    shellname = pwent->pw_shell;
+            }
+#endif
+        }
+        v = rb_str_new2(shellname);
+        argc = 1;
+        argv = &v;
     }
 
-    carg.execarg_obj = rb_execarg_new(argc, argv, 1);
+    carg.execarg_obj = rb_execarg_new(argc, argv, 1, 0);
     carg.eargp = rb_execarg_get(carg.execarg_obj);
     rb_execarg_parent_start(carg.execarg_obj);
 
@@ -193,13 +242,13 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
     pid = rb_fork_async_signal_safe(&status, chfunc, &carg, Qnil, errbuf, sizeof(errbuf));
 
     if (pid < 0) {
-	int e = errno;
-	close(master);
-	close(slave);
+        int e = errno;
+        close(master);
+        close(slave);
         rb_execarg_parent_end(carg.execarg_obj);
-	errno = e;
-	if (status) rb_jump_tag(status);
-	rb_sys_fail(errbuf[0] ? errbuf : "fork failed");
+        errno = e;
+        if (status) rb_jump_tag(status);
+        rb_sys_fail(errbuf[0] ? errbuf : "fork failed");
     }
 
     close(slave);
@@ -211,6 +260,21 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
     RB_GC_GUARD(carg.execarg_obj);
 }
 
+#if (defined(HAVE_POSIX_OPENPT) || defined(HAVE_PTSNAME)) && !defined(HAVE_PTSNAME_R)
+/* glibc only, not obsolete interface on Tru64 or HP-UX */
+static int
+ptsname_r(int fd, char *buf, size_t buflen)
+{
+    extern char *ptsname(int);
+    char *name = ptsname(fd);
+    if (!name) return -1;
+    strlcpy(buf, name, buflen);
+    return 0;
+}
+# define HAVE_PTSNAME_R 1
+#endif
+
+#if defined(HAVE_POSIX_OPENPT) || defined(HAVE_OPENPTY) || defined(HAVE_PTSNAME_R)
 static int
 no_mesg(char *slavedevice, int nomesg)
 {
@@ -219,6 +283,22 @@ no_mesg(char *slavedevice, int nomesg)
     else
         return 0;
 }
+#endif
+
+#if defined(I_PUSH) && !defined(__linux__) && !defined(_AIX)
+static inline int
+ioctl_I_PUSH(int fd, const char *const name)
+{
+    int ret = 0;
+# if defined(I_FIND)
+    ret = ioctl(fd, I_FIND, name);
+# endif
+    if (ret == 0) {
+        ret = ioctl(fd, I_PUSH, name);
+    }
+    return ret;
+}
+#endif
 
 static int
 get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg, int fail)
@@ -233,7 +313,7 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     dfl.sa_flags = 0;
     sigemptyset(&dfl.sa_mask);
 
-#if defined(__sun) || (defined(__FreeBSD__) && __FreeBSD_version < 902000)
+#if defined(__sun) || defined(__OpenBSD__) || (defined(__FreeBSD__) && __FreeBSD_version < 902000)
     /* workaround for Solaris 10: grantpt() doesn't work if FD_CLOEXEC is set.  [ruby-dev:44688] */
     /* FreeBSD 9.2 or later supports O_CLOEXEC
      * http://www.freebsd.org/cgi/query-pr.cgi?pr=162374 */
@@ -243,14 +323,14 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     rb_fd_fix_cloexec(masterfd);
 #else
     {
-	int flags = O_RDWR|O_NOCTTY;
+        int flags = O_RDWR|O_NOCTTY;
 # if defined(O_CLOEXEC)
-	/* glibc posix_openpt() in GNU/Linux calls open("/dev/ptmx", flags) internally.
-	 * So version dependency on GNU/Linux is same as O_CLOEXEC with open().
-	 * O_CLOEXEC is available since Linux 2.6.23.  Linux 2.6.18 silently ignore it. */
-	flags |= O_CLOEXEC;
+        /* glibc posix_openpt() in GNU/Linux calls open("/dev/ptmx", flags) internally.
+         * So version dependency on GNU/Linux is the same as O_CLOEXEC with open().
+         * O_CLOEXEC is available since Linux 2.6.23.  Linux 2.6.18 silently ignore it. */
+        flags |= O_CLOEXEC;
 # endif
-	if ((masterfd = posix_openpt(flags)) == -1) goto error;
+        if ((masterfd = posix_openpt(flags)) == -1) goto error;
     }
     rb_fd_fix_cloexec(masterfd);
     if (sigaction(SIGCHLD, &dfl, &old) == -1) goto error;
@@ -258,20 +338,20 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
 #endif
     if (sigaction(SIGCHLD, &old, NULL) == -1) goto error;
     if (unlockpt(masterfd) == -1) goto error;
-    if ((slavedevice = ptsname(masterfd)) == NULL) goto error;
+    if (ptsname_r(masterfd, SlaveName, DEVICELEN) != 0) goto error;
+    slavedevice = SlaveName;
     if (no_mesg(slavedevice, nomesg) == -1) goto error;
     if ((slavefd = rb_cloexec_open(slavedevice, O_RDWR|O_NOCTTY, 0)) == -1) goto error;
     rb_update_max_fd(slavefd);
 
 #if defined(I_PUSH) && !defined(__linux__) && !defined(_AIX)
-    if (ioctl(slavefd, I_PUSH, "ptem") == -1) goto error;
-    if (ioctl(slavefd, I_PUSH, "ldterm") == -1) goto error;
-    if (ioctl(slavefd, I_PUSH, "ttcompat") == -1) goto error;
+    if (ioctl_I_PUSH(slavefd, "ptem") == -1) goto error;
+    if (ioctl_I_PUSH(slavefd, "ldterm") == -1) goto error;
+    if (ioctl_I_PUSH(slavefd, "ttcompat") == -1) goto error;
 #endif
 
     *master = masterfd;
     *slave = slavefd;
-    strlcpy(SlaveName, slavedevice, DEVICELEN);
     return 0;
 
   grantpt_error:
@@ -289,15 +369,15 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
  * or the same interface function.
  */
     if (openpty(master, slave, SlaveName,
-		(struct termios *)0, (struct winsize *)0) == -1) {
-	if (!fail) return -1;
-	rb_raise(rb_eRuntimeError, "openpty() failed");
+                (struct termios *)0, (struct winsize *)0) == -1) {
+        if (!fail) return -1;
+        rb_raise(rb_eRuntimeError, "openpty() failed");
     }
     rb_fd_fix_cloexec(*master);
     rb_fd_fix_cloexec(*slave);
     if (no_mesg(SlaveName, nomesg) == -1) {
-	if (!fail) return -1;
-	rb_raise(rb_eRuntimeError, "can't chmod slave pty");
+        if (!fail) return -1;
+        rb_raise(rb_eRuntimeError, "can't chmod slave pty");
     }
 
     return 0;
@@ -308,8 +388,8 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     mode_t mode = nomesg ? 0600 : 0622;
 
     if (!(name = _getpty(master, O_RDWR, mode, 0))) {
-	if (!fail) return -1;
-	rb_raise(rb_eRuntimeError, "_getpty() failed");
+        if (!fail) return -1;
+        rb_raise(rb_eRuntimeError, "_getpty() failed");
     }
     rb_fd_fix_cloexec(*master);
 
@@ -325,7 +405,6 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     char *slavedevice;
     void (*s)();
 
-    extern char *ptsname(int);
     extern int unlockpt(int);
     extern int grantpt(int);
 
@@ -343,18 +422,18 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
 #endif
     signal(SIGCHLD, s);
     if(unlockpt(masterfd) == -1) goto error;
-    if((slavedevice = ptsname(masterfd)) == NULL) goto error;
+    if (ptsname_r(masterfd, SlaveName, DEVICELEN) != 0) goto error;
+    slavedevice = SlaveName;
     if (no_mesg(slavedevice, nomesg) == -1) goto error;
     if((slavefd = rb_cloexec_open(slavedevice, O_RDWR, 0)) == -1) goto error;
     rb_update_max_fd(slavefd);
 #if defined(I_PUSH) && !defined(__linux__) && !defined(_AIX)
-    if(ioctl(slavefd, I_PUSH, "ptem") == -1) goto error;
-    if(ioctl(slavefd, I_PUSH, "ldterm") == -1) goto error;
-    ioctl(slavefd, I_PUSH, "ttcompat");
+    if(ioctl_I_PUSH(slavefd, "ptem") == -1) goto error;
+    if(ioctl_I_PUSH(slavefd, "ldterm") == -1) goto error;
+    ioctl_I_PUSH(slavefd, "ttcompat");
 #endif
     *master = masterfd;
     *slave = slavefd;
-    strlcpy(SlaveName, slavedevice, DEVICELEN);
     return 0;
 
   error:
@@ -365,83 +444,46 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
 #else
     /* BSD */
     int	 masterfd = -1, slavefd = -1;
-    const char *const *p;
+    int  i;
     char MasterName[DEVICELEN];
 
-#if defined(__hpux)
-    static const char MasterDevice[] = "/dev/ptym/pty%s";
-    static const char SlaveDevice[] =  "/dev/pty/tty%s";
-    static const char *const deviceNo[] = {
-    "p0","p1","p2","p3","p4","p5","p6","p7",
-    "p8","p9","pa","pb","pc","pd","pe","pf",
-    "q0","q1","q2","q3","q4","q5","q6","q7",
-    "q8","q9","qa","qb","qc","qd","qe","qf",
-    "r0","r1","r2","r3","r4","r5","r6","r7",
-    "r8","r9","ra","rb","rc","rd","re","rf",
-    "s0","s1","s2","s3","s4","s5","s6","s7",
-    "s8","s9","sa","sb","sc","sd","se","sf",
-    "t0","t1","t2","t3","t4","t5","t6","t7",
-    "t8","t9","ta","tb","tc","td","te","tf",
-    "u0","u1","u2","u3","u4","u5","u6","u7",
-    "u8","u9","ua","ub","uc","ud","ue","uf",
-    "v0","v1","v2","v3","v4","v5","v6","v7",
-    "v8","v9","va","vb","vc","vd","ve","vf",
-    "w0","w1","w2","w3","w4","w5","w6","w7",
-    "w8","w9","wa","wb","wc","wd","we","wf",
-    0
-    };
-#elif defined(_IBMESA)  /* AIX/ESA */
+#define HEX1(c) \
+        c"0",c"1",c"2",c"3",c"4",c"5",c"6",c"7", \
+        c"8",c"9",c"a",c"b",c"c",c"d",c"e",c"f"
+
+#if defined(_IBMESA)  /* AIX/ESA */
     static const char MasterDevice[] = "/dev/ptyp%s";
     static const char SlaveDevice[] = "/dev/ttyp%s";
-    static const char *const deviceNo[] = {
-    "00","01","02","03","04","05","06","07","08","09","0a","0b","0c","0d","0e","0f",
-    "10","11","12","13","14","15","16","17","18","19","1a","1b","1c","1d","1e","1f",
-    "20","21","22","23","24","25","26","27","28","29","2a","2b","2c","2d","2e","2f",
-    "30","31","32","33","34","35","36","37","38","39","3a","3b","3c","3d","3e","3f",
-    "40","41","42","43","44","45","46","47","48","49","4a","4b","4c","4d","4e","4f",
-    "50","51","52","53","54","55","56","57","58","59","5a","5b","5c","5d","5e","5f",
-    "60","61","62","63","64","65","66","67","68","69","6a","6b","6c","6d","6e","6f",
-    "70","71","72","73","74","75","76","77","78","79","7a","7b","7c","7d","7e","7f",
-    "80","81","82","83","84","85","86","87","88","89","8a","8b","8c","8d","8e","8f",
-    "90","91","92","93","94","95","96","97","98","99","9a","9b","9c","9d","9e","9f",
-    "a0","a1","a2","a3","a4","a5","a6","a7","a8","a9","aa","ab","ac","ad","ae","af",
-    "b0","b1","b2","b3","b4","b5","b6","b7","b8","b9","ba","bb","bc","bd","be","bf",
-    "c0","c1","c2","c3","c4","c5","c6","c7","c8","c9","ca","cb","cc","cd","ce","cf",
-    "d0","d1","d2","d3","d4","d5","d6","d7","d8","d9","da","db","dc","dd","de","df",
-    "e0","e1","e2","e3","e4","e5","e6","e7","e8","e9","ea","eb","ec","ed","ee","ef",
-    "f0","f1","f2","f3","f4","f5","f6","f7","f8","f9","fa","fb","fc","fd","fe","ff",
-    0
+    static const char deviceNo[][3] = {
+        HEX1("0"), HEX1("1"), HEX1("2"), HEX1("3"),
+        HEX1("4"), HEX1("5"), HEX1("6"), HEX1("7"),
+        HEX1("8"), HEX1("9"), HEX1("a"), HEX1("b"),
+        HEX1("c"), HEX1("d"), HEX1("e"), HEX1("f"),
     };
 #else /* 4.2BSD */
     static const char MasterDevice[] = "/dev/pty%s";
     static const char SlaveDevice[] = "/dev/tty%s";
-    static const char *const deviceNo[] = {
-    "p0","p1","p2","p3","p4","p5","p6","p7",
-    "p8","p9","pa","pb","pc","pd","pe","pf",
-    "q0","q1","q2","q3","q4","q5","q6","q7",
-    "q8","q9","qa","qb","qc","qd","qe","qf",
-    "r0","r1","r2","r3","r4","r5","r6","r7",
-    "r8","r9","ra","rb","rc","rd","re","rf",
-    "s0","s1","s2","s3","s4","s5","s6","s7",
-    "s8","s9","sa","sb","sc","sd","se","sf",
-    0
+    static const char deviceNo[][3] = {
+        HEX1("p"), HEX1("q"), HEX1("r"), HEX1("s"),
     };
 #endif
-    for (p = deviceNo; *p != NULL; p++) {
-	snprintf(MasterName, sizeof MasterName, MasterDevice, *p);
-	if ((masterfd = rb_cloexec_open(MasterName,O_RDWR,0)) >= 0) {
+#undef HEX1
+    for (i = 0; i < numberof(deviceNo); i++) {
+        const char *const devno = deviceNo[i];
+        snprintf(MasterName, sizeof MasterName, MasterDevice, devno);
+        if ((masterfd = rb_cloexec_open(MasterName,O_RDWR,0)) >= 0) {
             rb_update_max_fd(masterfd);
-	    *master = masterfd;
-	    snprintf(SlaveName, DEVICELEN, SlaveDevice, *p);
-	    if ((slavefd = rb_cloexec_open(SlaveName,O_RDWR,0)) >= 0) {
+            *master = masterfd;
+            snprintf(SlaveName, DEVICELEN, SlaveDevice, devno);
+            if ((slavefd = rb_cloexec_open(SlaveName,O_RDWR,0)) >= 0) {
                 rb_update_max_fd(slavefd);
-		*slave = slavefd;
-		if (chown(SlaveName, getuid(), getgid()) != 0) goto error;
-		if (chmod(SlaveName, nomesg ? 0600 : 0622) != 0) goto error;
-		return 0;
-	    }
-	    close(masterfd);
-	}
+                *slave = slavefd;
+                if (chown(SlaveName, getuid(), getgid()) != 0) goto error;
+                if (chmod(SlaveName, nomesg ? 0600 : 0622) != 0) goto error;
+                return 0;
+            }
+            close(masterfd);
+        }
     }
   error:
     if (slavefd != -1) close(slavefd);
@@ -455,8 +497,8 @@ static void
 getDevice(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg)
 {
     if (get_device_once(master, slave, SlaveName, nomesg, 0)) {
-	rb_gc();
-	get_device_once(master, slave, SlaveName, nomesg, 1);
+        rb_gc();
+        get_device_once(master, slave, SlaveName, nomesg, 1);
     }
 }
 
@@ -468,8 +510,10 @@ pty_close_pty(VALUE assoc)
 
     for (i = 0; i < 2; i++) {
         io = rb_ary_entry(assoc, i);
-        if (RB_TYPE_P(io, T_FILE) && 0 <= RFILE(io)->fptr->fd)
+        if (RB_TYPE_P(io, T_FILE)) {
+            /* it's OK to call rb_io_close again even if it's already closed */
             rb_io_close(io);
+        }
     }
     return Qnil;
 }
@@ -477,11 +521,11 @@ pty_close_pty(VALUE assoc)
 /*
  * call-seq:
  *   PTY.open => [master_io, slave_file]
- *   PTY.open {|master_io, slave_file| ... } => block value
+ *   PTY.open {|(master_io, slave_file)| ... } => block value
  *
  * Allocates a pty (pseudo-terminal).
  *
- * In the block form, yields two arguments <tt>master_io, slave_file</tt>
+ * In the block form, yields an array of two elements (<tt>master_io, slave_file</tt>)
  * and the value of the block is returned from +open+.
  *
  * The IO and File are both closed after the block completes if they haven't
@@ -510,7 +554,7 @@ pty_close_pty(VALUE assoc)
  *   require 'io/console'
  *   PTY.open {|m, s|
  *     s.raw!
- *     ...
+ *     # ...
  *   }
  *
  */
@@ -519,38 +563,32 @@ pty_open(VALUE klass)
 {
     int master_fd, slave_fd;
     char slavename[DEVICELEN];
-    VALUE master_io, slave_file;
-    rb_io_t *master_fptr, *slave_fptr;
-    VALUE assoc;
 
     getDevice(&master_fd, &slave_fd, slavename, 1);
 
-    master_io = rb_obj_alloc(rb_cIO);
-    MakeOpenFile(master_io, master_fptr);
-    master_fptr->mode = FMODE_READWRITE | FMODE_SYNC | FMODE_DUPLEX;
-    master_fptr->fd = master_fd;
-    master_fptr->pathv = rb_obj_freeze(rb_sprintf("masterpty:%s", slavename));
+    VALUE master_path = rb_obj_freeze(rb_sprintf("masterpty:%s", slavename));
+    VALUE master_io = rb_io_open_descriptor(rb_cIO, master_fd, FMODE_READWRITE | FMODE_SYNC | FMODE_DUPLEX, master_path, RUBY_IO_TIMEOUT_DEFAULT, NULL);
 
-    slave_file = rb_obj_alloc(rb_cFile);
-    MakeOpenFile(slave_file, slave_fptr);
-    slave_fptr->mode = FMODE_READWRITE | FMODE_SYNC | FMODE_DUPLEX | FMODE_TTY;
-    slave_fptr->fd = slave_fd;
-    slave_fptr->pathv = rb_obj_freeze(rb_str_new_cstr(slavename));
+    VALUE slave_path = rb_obj_freeze(rb_str_new_cstr(slavename));
+    VALUE slave_file = rb_io_open_descriptor(rb_cFile, slave_fd, FMODE_READWRITE | FMODE_SYNC | FMODE_DUPLEX | FMODE_TTY, slave_path, RUBY_IO_TIMEOUT_DEFAULT, NULL);
+
+    VALUE assoc = rb_assoc_new(master_io, slave_file);
 
-    assoc = rb_assoc_new(master_io, slave_file);
     if (rb_block_given_p()) {
-	return rb_ensure(rb_yield, assoc, pty_close_pty, assoc);
+        return rb_ensure(rb_yield, assoc, pty_close_pty, assoc);
     }
+
     return assoc;
 }
 
 static VALUE
-pty_detach_process(struct pty_info *info)
+pty_detach_process(VALUE v)
 {
+    struct pty_info *info = (void *)v;
 #ifdef WNOHANG
     int st;
     if (rb_waitpid(info->child_pid, &st, WNOHANG) <= 0)
-	return Qnil;
+        return Qnil;
 #endif
     rb_detach_process(info->child_pid);
     return Qnil;
@@ -558,10 +596,10 @@ pty_detach_process(struct pty_info *info)
 
 /*
  * call-seq:
- *   PTY.spawn(command_line)  { |r, w, pid| ... }
- *   PTY.spawn(command_line)  => [r, w, pid]
- *   PTY.spawn(command, arguments, ...)  { |r, w, pid| ... }
- *   PTY.spawn(command, arguments, ...)  => [r, w, pid]
+ *   PTY.spawn([env,] command_line)  { |r, w, pid| ... }
+ *   PTY.spawn([env,] command_line)  => [r, w, pid]
+ *   PTY.spawn([env,] command, arguments, ...)  { |r, w, pid| ... }
+ *   PTY.spawn([env,] command, arguments, ...)  => [r, w, pid]
  *
  * Spawns the specified command on a newly allocated pty. You can also use the
  * alias ::getpty.
@@ -569,6 +607,21 @@ pty_detach_process(struct pty_info *info)
  * The command's controlling tty is set to the slave device of the pty
  * and its standard input/output/error is redirected to the slave device.
  *
+ * +env+ is an optional hash that provides additional environment variables to the spawned pty.
+ *
+ *   # sets FOO to "bar"
+ *   PTY.spawn({"FOO"=>"bar"}, "printenv", "FOO") do |r, w, pid|
+ *     p r.read #=> "bar\r\n"
+ *   ensure
+ *     r.close; w.close; Process.wait(pid)
+ *   end
+ *   # unsets FOO
+ *   PTY.spawn({"FOO"=>nil}, "printenv", "FOO") do |r, w, pid|
+ *     p r.read #=> ""
+ *   ensure
+ *     r.close; w.close; Process.wait(pid)
+ *   end
+ *
  * +command+ and +command_line+ are the full commands to run, given a String.
  * Any additional +arguments+ will be passed to the command.
  *
@@ -579,45 +632,51 @@ pty_detach_process(struct pty_info *info)
  *
  * In the block form these same values will be yielded to the block:
  *
- * +r+:: A readable IO that that contains the command's
+ * +r+:: A readable IO that contains the command's
  *       standard output and standard error
  * +w+:: A writable IO that is the command's standard input
  * +pid+:: The process identifier for the command.
+ *
+ * === Clean up
+ *
+ * This method does not clean up like closing IOs or waiting for child
+ * process, except that the process is detached in the block form to
+ * prevent it from becoming a zombie (see Process.detach).  Any other
+ * cleanup is the responsibility of the caller.  If waiting for +pid+,
+ * be sure to close both +r+ and +w+ before doing so; doing it in the
+ * reverse order may cause deadlock on some OSes.
  */
 static VALUE
 pty_getpty(int argc, VALUE *argv, VALUE self)
 {
     VALUE res;
     struct pty_info info;
-    rb_io_t *wfptr,*rfptr;
-    VALUE rport = rb_obj_alloc(rb_cFile);
-    VALUE wport = rb_obj_alloc(rb_cFile);
     char SlaveName[DEVICELEN];
 
-    MakeOpenFile(rport, rfptr);
-    MakeOpenFile(wport, wfptr);
-
     establishShell(argc, argv, &info, SlaveName);
 
-    rfptr->mode = rb_io_modestr_fmode("r");
-    rfptr->fd = info.fd;
-    rfptr->pathv = rb_obj_freeze(rb_str_new_cstr(SlaveName));
+    VALUE pty_path = rb_obj_freeze(rb_str_new_cstr(SlaveName));
+    VALUE rport = rb_io_open_descriptor(
+        rb_cFile, info.fd, FMODE_READABLE, pty_path, RUBY_IO_TIMEOUT_DEFAULT, NULL
+    );
 
-    wfptr->mode = rb_io_modestr_fmode("w") | FMODE_SYNC;
-    wfptr->fd = rb_cloexec_dup(info.fd);
-    if (wfptr->fd == -1)
+    int wpty_fd = rb_cloexec_dup(info.fd);
+    if (wpty_fd == -1) {
         rb_sys_fail("dup()");
-    rb_update_max_fd(wfptr->fd);
-    wfptr->pathv = rfptr->pathv;
+    }
+    VALUE wport = rb_io_open_descriptor(
+        rb_cFile, wpty_fd, FMODE_WRITABLE | FMODE_TRUNC | FMODE_CREATE | FMODE_SYNC,
+        pty_path, RUBY_IO_TIMEOUT_DEFAULT, NULL
+    );
 
     res = rb_ary_new2(3);
-    rb_ary_store(res,0,(VALUE)rport);
-    rb_ary_store(res,1,(VALUE)wport);
+    rb_ary_store(res, 0, rport);
+    rb_ary_store(res, 1, wport);
     rb_ary_store(res,2,PIDT2NUM(info.child_pid));
 
     if (rb_block_given_p()) {
-	rb_ensure(rb_yield, res, pty_detach_process, (VALUE)&info);
-	return Qnil;
+        rb_ensure(rb_yield, res, pty_detach_process, (VALUE)&info);
+        return Qnil;
     }
     return res;
 }
@@ -637,13 +696,13 @@ raise_from_check(rb_pid_t pid, int status)
 ---->> Either IF_STOPPED or WIFSTOPPED is needed <<----
 #endif /* WIFSTOPPED | IF_STOPPED */
     if (WIFSTOPPED(status)) { /* suspend */
-	state = "stopped";
+        state = "stopped";
     }
     else if (kill(pid, 0) == 0) {
-	state = "changed";
+        state = "changed";
     }
     else {
-	state = "exited";
+        state = "exited";
     }
     msg = rb_sprintf("pty - %s: %ld", state, (long)pid);
     exc = rb_exc_new_str(eChildExited, msg);
@@ -674,15 +733,23 @@ pty_check(int argc, VALUE *argv, VALUE self)
     VALUE pid, exc;
     rb_pid_t cpid;
     int status;
+    const int flag =
+#ifdef WNOHANG
+        WNOHANG|
+#endif
+#ifdef WUNTRACED
+        WUNTRACED|
+#endif
+        0;
 
     rb_scan_args(argc, argv, "11", &pid, &exc);
-    cpid = rb_waitpid(NUM2PIDT(pid), &status, WNOHANG|WUNTRACED);
+    cpid = rb_waitpid(NUM2PIDT(pid), &status, flag);
     if (cpid == -1 || cpid == 0) return Qnil;
 
     if (!RTEST(exc)) return rb_last_status_get();
     raise_from_check(cpid, status);
 
-    UNREACHABLE;
+    UNREACHABLE_RETURN(Qnil);
 }
 
 static VALUE cPTY;
@@ -697,8 +764,8 @@ static VALUE cPTY;
 /*
  * Document-class: PTY
  *
- * Creates and managed pseudo terminals (PTYs).  See also
- * http://en.wikipedia.org/wiki/Pseudo_terminal
+ * Creates and manages pseudo terminals (PTYs).  See also
+ * https://en.wikipedia.org/wiki/Pseudo_terminal
  *
  * PTY allows you to allocate new terminals using ::open or ::spawn a new
  * terminal with a specific command.
@@ -741,24 +808,29 @@ static VALUE cPTY;
  *
  * == License
  *
- *  C) Copyright 1998 by Akinori Ito.
+ * (c) Copyright 1998 by Akinori Ito.
  *
- *  This software may be redistributed freely for this purpose, in full
- *  or in part, provided that this entire copyright notice is included
- *  on any copies of this software and applications and derivations thereof.
+ * This software may be redistributed freely for this purpose, in full
+ * or in part, provided that this entire copyright notice is included
+ * on any copies of this software and applications and derivations thereof.
  *
- *  This software is provided on an "as is" basis, without warranty of any
- *  kind, either expressed or implied, as to any matter including, but not
- *  limited to warranty of fitness of purpose, or merchantability, or
- *  results obtained from use of this software.
+ * This software is provided on an "as is" basis, without warranty of any
+ * kind, either expressed or implied, as to any matter including, but not
+ * limited to warranty of fitness of purpose, or merchantability, or
+ * results obtained from use of this software.
  */
 
 void
 Init_pty(void)
 {
     cPTY = rb_define_module("PTY");
-    /* :nodoc: */
-    rb_define_module_function(cPTY,"getpty",pty_getpty,-1);
+#if 1
+    rb_define_module_function(cPTY,"get""pty",pty_getpty,-1);
+#else  /* for RDoc */
+    /* show getpty as an alias of spawn */
+    VALUE sPTY = rb_singleton_class(cPTY);
+    rb_define_alias(sPTY, "getpty", "spawn");
+#endif
     rb_define_module_function(cPTY,"spawn",pty_getpty,-1);
     rb_define_singleton_method(cPTY,"check",pty_check,-1);
     rb_define_singleton_method(cPTY,"open",pty_open,0);
